/*
    This file is part of Konsole, an X terminal.

    Copyright 2007-2008 by Robert Knight <robertknight@gmail.com>
    Copyright 1997,1998 by Lars Doelle <lars.doelle@on-line.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301  USA.
*/

#ifndef EMULATION_H
#define EMULATION_H

// System
#include <cstdio>

// Qt
#include <QKeyEvent>
//#include <QPointer>
#include <QTextCodec>
#include <QTextStream>
#include <QTimer>

#include "qtermwidget_export.h"

namespace Konsole {

class KeyboardTranslator;
class HistoryType;
class Screen;
class ScreenWindow;
class TerminalCharacterDecoder;

/**
 * This enum describes the available states which
 * the terminal emulation may be set to.
 *
 * These are the values used by Emulation::stateChanged()
 */
enum {
    /** The emulation is currently receiving user input. */
    NOTIFYNORMAL = 0,
    /**
     * The terminal program has triggered a bell event
     * to get the user's attention.
     */
    NOTIFYBELL = 1,
    /**
     * The emulation is currently receiving data from its
     * terminal input.
     */
    NOTIFYACTIVITY = 2,

    // unused here?
    NOTIFYSILENCE = 3
};

/**
 * Base class for terminal emulation back-ends.
 *
 * The back-end is responsible for decoding an incoming character stream and
 * producing an output image of characters.
 *
 * When input from the terminal is received, the receiveData() slot should be called with
 * the data which has arrived.  The emulation will process the data and update the
 * screen image accordingly.  The codec used to decode the incoming character stream
 * into the unicode characters used internally can be specified using setCodec()
 *
 * The size of the screen image can be specified by calling setImageSize() with the
 * desired number of lines and columns.  When new lines are added, old content
 * is moved into a history store, which can be set by calling setHistory().
 *
 * The screen image can be accessed by creating a ScreenWindow onto this emulation
 * by calling createWindow().  Screen windows provide access to a section of the
 * output.  Each screen window covers the same number of lines and columns as the
 * image size returned by imageSize().  The screen window can be moved up and down
 * and provides transparent access to both the current on-screen image and the
 * previous output.  The screen windows emit an outputChanged signal
 * when the section of the image they are looking at changes.
 * Graphical views can then render the contents of a screen window, listening for notifications
 * of output changes from the screen window which they are associated with and updating
 * accordingly.
 *
 * The emulation also is also responsible for converting input from the connected views such
 * as keypresses and mouse activity into a character string which can be sent
 * to the terminal program.  Key presses can be processed by calling the sendKeyEvent() slot,
 * while mouse events can be processed using the sendMouseEvent() slot.  When the character
 * stream has been produced, the emulation will emit a sendData() signal with a pointer
 * to the character buffer.  This data should be fed to the standard input of the terminal
 * process.  The translation of key presses into an output character stream is performed
 * using a lookup in a set of key bindings which map key sequences to output
 * character sequences.  The name of the key bindings set used can be specified using
 * setKeyBindings()
 *
 * The emulation maintains certain state information which changes depending on the
 * input received.  The emulation can be reset back to its starting state by calling
 * reset().
 *
 * The emulation also maintains an activity state, which specifies whether
 * terminal is currently active ( when data is received ), normal
 * ( when the terminal is idle or receiving user input ) or trying
 * to alert the user ( also known as a "Bell" event ).  The stateSet() signal
 * is emitted whenever the activity state is set.  This can be used to determine
 * how long the emulation has been active/idle for and also respond to
 * a 'bell' event in different ways.
 */
class TERMINALWIDGET_EXPORT Emulation : public QObject
{
    Q_OBJECT

public:

    /**
     * This enum describes the available shapes for the keyboard cursor.
     * See setKeyboardCursorShape()
     */
    enum class KeyboardCursorShape {
        /** A rectangular block which covers the entire area of the cursor character. */
        BlockCursor = 0,
        /**
         * A single flat line which occupies the space at the bottom of the cursor
         * character's area.
         */
        UnderlineCursor = 1,
        /**
         * An cursor shaped like the capital letter 'I', similar to the IBeam
         * cursor used in Qt/KDE text editors.
         */
        IBeamCursor = 2
    };

    Emulation();
    ~Emulation() override;

    ScreenWindow *createWindow();

    QSize imageSize() const;

    int lineCount() const;

    int columnCount() const;

    void setHistory(const HistoryType &);

    const HistoryType &history() const;

    void clearHistory();

    virtual void writeToStream(TerminalCharacterDecoder *decoder, int startLine, int endLine);

    const QTextCodec *codec() const
    {
        return _codec;
    }

    void setCodec(const QTextCodec *);

    bool utf8() const
    {
        Q_ASSERT(_codec);
        return _codec->mibEnum() == 106;
    }

    virtual char eraseChar() const;

    void setKeyBindings(const QString &name);

    QString keyBindings() const;

    virtual void clearEntireScreen() = 0;

    virtual void reset() = 0;

    bool programUsesMouse() const;

    bool programBracketedPasteMode() const;

    void setKeyTranslator(const KeyboardTranslator *keyTranslator);

    void setBackspaceMode(char *key, int length);
    void setDeleteMode(char *key, int length);

    void setSessionId(int sessionId);

public slots:

    virtual void setImageSize(int lines, int columns);

    virtual void sendText(const QString &text) = 0;

    virtual void sendKeyEvent(QKeyEvent *);

    virtual void sendMouseEvent(int buttons, int column, int line, int eventType);

    virtual void sendString(const char *string, int length = -1) = 0;

    void receiveData(const char *buffer, int len, bool isCommandExec);

signals:

    void sendData(const char *data, int len, const QTextCodec *codec);

    void lockPtyRequest(bool suspend);

    void useUtf8Request(bool);

    void stateSet(int state);

    void zmodemDetected();

    void changeTabTextColorRequest(int color);

    void programUsesMouseChanged(bool usesMouse);

    void programBracketedPasteModeChanged(bool bracketedPasteMode);

    void outputChanged();

    void titleChanged(int title, const QString &newTitle);

    void imageSizeChanged(int lineCount, int columnCount);

    void imageSizeInitialized();

    void imageResizeRequest(const QSize &sizz);

    void profileChangeCommandReceived(const QString &text);

    void flowControlKeyPressed(bool suspendKeyPressed);

    void cursorChanged(KeyboardCursorShape cursorShape, bool blinkingCursorEnabled);

protected:
    virtual void setMode(int mode) = 0;
    virtual void resetMode(int mode) = 0;

    virtual void receiveChar(wchar_t ch);

    void setScreen(int index);

    enum EmulationCodec {
        LocaleCodec = 0,
        Utf8Codec   = 1
    };
    void setCodec(EmulationCodec codec);


    QList<ScreenWindow *> _windows;

    Screen *_currentScreen;

    Screen *_screen[2];
    const QTextCodec *_codec;
    QTextDecoder *_decoder;

    KeyboardTranslator *_keyTranslator;

protected slots:
    void bufferedUpdate();

private slots:
    void showBulk();

    void usesMouseChanged(bool usesMouse);

    void bracketedPasteModeChanged(bool bracketedPasteMode);

private:
    bool _usesMouse;
    bool _bracketedPasteMode;
    QTimer _bulkTimer1;
    QTimer _bulkTimer2;

    int _sessionId;
};

}

#endif // ifndef EMULATION_H
